"""
"""


#[
from __future__ import absolute_import

from typing import (Sequence, )
import plotly.graph_objects as _pg
import plotly.subplots as _ps
#]


__all__ = (
    "make_subplots",
)


_EMPTY_SUBPLOT_TITLE = " "


def make_subplots(
    rows: int,
    columns: int | None = None,
    cols: int | None = None,
    subplot_titles: Sequence[str] | bool | None = None,
    **kwargs,
) -> _pg.Figure:
    """
    """
    if cols is not None and columns is None:
        columns = cols
    total_num_subplots = rows * columns
    if subplot_titles is True:
        subplot_titles = (_EMPTY_SUBPLOT_TITLE, )*total_num_subplots
    return _ps.make_subplots(
        rows=rows,
        cols=columns,
        subplot_titles=subplot_titles or None,
        **kwargs,
    )




r"""
# First-order solution matrices


## Square solution:

$$
\begin{gathered}
\xi_t = T \, \xi_{t-1} + R \, v_t + K
\\
y_t = Z \, \xi_t + H \, w_t + D
\end{gathered}
$$


## Equivalent block-triangular solution:

$$
\begin{gathered}
\alpha_t = T_\alpha \, \alpha_{t-1} + R_\alpha \, v_t + K_\alpha
\\
y_t = Z_\alpha \, \alpha_t + D + H \, w_t
\\
\xi_t \equiv = U_alpha \, \alpha_t
\end{gathered}
$$


## Forward expansion:

$$
$$
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )
from numbers import (Real, )
import dataclasses as _dc

import enum as _en
import numpy as _np
import scipy as _sp

from ..fords import systems as _systems
from ..fords import descriptors as _descriptors
#]


class EigenValueKind(_en.Flag):
    STABLE = _en.auto()
    UNIT = _en.auto()
    UNSTABLE = _en.auto()


class SystemStabilityKind(_en.Flag):
    STABLE = _en.auto()
    MULTIPLE_STABLE = _en.auto()
    NO_STABLE = _en.auto()


class VariableStability(_en.Flag):
    STABLE = _en.auto()
    UNIT_ROOT = _en.auto()


@_dc.dataclass(slots=True, )
class Solution:
    """
    ## Square solution:

    T: Transition matrix
    R: Impact matrix of transition shocks
    K: Intercept in transition equation
    Z: Measurement matrix
    H: Impact matrix of measurement shocks
    D: Intercept in measurement equation


    ## Equivalent block-triangular solution:

    Ta: Transition matrix in triangular system
    Ra: Impact matrix of transition shocks in triangular system
    Ka: Intercept in transition equation in triangular system
    Za: Measurement matrix in triangular system
    Ua: Rotation matrix from triangular to square system


    ## Forward expansion of square solution:

    J: Power matrix
    Ru: Forward-looking impact matrix of transition shocks
    X: Impact matrix in square system
    Xa: Impact matrix in triangular system
    """
    #[
    T: _np.ndarray | None = None
    R: _np.ndarray | None = None
    K: _np.ndarray | None = None
    Z: _np.ndarray | None = None
    H: _np.ndarray | None = None
    D: _np.ndarray | None = None

    Ta: _np.ndarray | None = None
    Ra: _np.ndarray | None = None
    Ka: _np.ndarray | None = None
    Za: _np.ndarray | None = None
    Ua: _np.ndarray | None = None

    J: _np.ndarray | None = None
    Ru: _np.ndarray | None = None
    X: _np.ndarray | None = None
    Xa: _np.ndarray | None = None

    eigen_values: tuple[Real, ...] | None = None
    eigen_values_stability: tuple[EigenValueKind, ...] | None = None
    system_stability: SystemStabilityKind | None = None
    transition_vector_stability: VariableStability | None = None
    measurement_vector_stability: VariableStability | None = None

    def __init__(
        self, 
        descriptor: _descriptors.Descriptor,
        system: _systems.System,
        /,
        *,
        tolerance: float = 1e-12,
    ) -> Self:
        """
        """
        def is_alpha_beta_stable_or_unit_root(alpha: Real, beta: Real, /, ) -> bool:
            return abs(beta) < (1 + tolerance)*abs(alpha)
        def is_stable_root(root: Real, /, ) -> bool:
            return abs(root) < (1 - tolerance)
        def is_unit_root(root: Real, /, ) -> bool:
            return abs(root) >= (1 - tolerance) and abs(root) < (1 + tolerance)
        #
        # Detach unstable from (stable + unit) roots and solve out expectations
        # The system is triangular but because stable and unit roots are
        # not detached yet, the system is called "preliminary"
        qz, self.eigen_values = \
            _solve_ordqz(system, is_alpha_beta_stable_or_unit_root, is_stable_root, is_unit_root, )
        triangular_solution_prelim = \
            _solve_transition_equations(descriptor, system, qz, )
        #
        # Detach unit from stable roots to create the final triangular solution
        # From the final triangular solution, calculate the square solution
        triangular_solution = \
            detach_stable_from_unit_roots(triangular_solution_prelim, is_unit_root, )
        square_solution = \
            _square_from_triangular(triangular_solution, )
        self.Ua, self.Ta, self.Ra, self.Ka, self.Xa, self.J, self.Ru = triangular_solution
        self.T, self.R, self.K, self.X = square_solution
        #
        # Solve measurement equations
        self.Z, self.H, self.D, self.Za = _solve_measurement_equations(descriptor, system, self.Ua, )
        self._classify_eigen_values_stability(is_stable_root, is_unit_root, )
        self._classify_system_stability(descriptor.get_num_forwards(), )
        self._classify_transition_vector_stability(tolerance=tolerance, )
        self._classify_measurement_vector_stability(tolerance=tolerance, )

    @property
    def num_alpha(self, /, ) -> int:
        """
        Number of alpha vector elements
        """
        return self.T.shape[0]

    @property
    def num_y(self, /, ) -> int:
        """
        Number of y vector elements
        """
        return self.Z.shape[0]

    @property
    def num_v(self, /, ) -> int:
        """
        Number of v vector elements
        """
        return self.R.shape[1]

    @property
    def num_w(self, /, ) -> int:
        """
        Number of w vector elements
        """
        return self.H.shape[1]

    @property
    def num_unit_roots(self, /, ) -> int:
        """
        Number of unit roots
        """
        return self.eigen_values_stability.count(EigenValueKind.UNIT)

    @property
    def Ta_stable(self, /, ) -> _np.ndarray:
        """
        Stable part of transition matrix
        """
        num_unit_roots = self.num_unit_roots
        return self.Ta[num_unit_roots:, num_unit_roots:]

    @property
    def Ra_stable(self, /, ) -> _np.ndarray:
        """
        Stable part of impact matrix of transition shocks
        """
        num_unit_roots = self.num_unit_roots
        return self.Ra[num_unit_roots:, :]

    @property
    def Za_stable(self, /, ) -> _np.ndarray:
        """
        Stable part of measurement matrix
        """
        num_unit_roots = self.num_unit_roots
        return self.Za[:, num_unit_roots:]

    @property
    def boolex_stable_transition_vector(self, /, ) -> tuple[int, ...]:
        """
        Index of stable transition vector elements
        """
        return _np.array(tuple(
            i == VariableStability.STABLE
            for i in self.transition_vector_stability
        ), dtype=bool, )

    @property
    def boolex_stable_measurement_vector(self, /, ) -> tuple[int, ...]:
        """
        Index of stable measurement vector elements
        """
        return _np.array(tuple(
            i == VariableStability.STABLE
            for i in self.measurement_vector_stability
        ), dtype=bool, )

    def expand_square_solution(self, forward, /, ) -> list[_np.ndarray]:
        """
        Expand R matrices of square solution for t+1...t+forward
        """
        R, X, J, Ru = self.R, self.X, self.J, self.Ru
        if (R is None) or (X is None) or (J is None) or (Ru is None):
            return None
        #
        # return [R(t+1), R(t+2), ..., R(t+forward)]
        #
        # R(t+k) = -X J**(k-1) Ru e(t+k)
        # k = 1, ..., forward or k-1 = 0, ..., forward-1
        #
        return [
            -X @ _np.linalg.matrix_power(J, k_minus_1) @ Ru 
            for k_minus_1 in range(0, forward)
        ]

    def _classify_eigen_values_stability(
        self,
        is_stable_root: Callable[[Real], bool],
        is_unit_root: Callable[[Real], bool],
        /,
    ) -> None:
        self.eigen_values_stability = tuple(
            _classify_eigen_value_stability(v, is_stable_root, is_unit_root, )
            for v in self.eigen_values
        )

    def _classify_system_stability(
        self,
        num_forwards: int,
        /,
    ) -> None:
        num_unstable = self.eigen_values_stability.count(EigenValueKind.UNSTABLE)
        if num_unstable == num_forwards:
            self.system_stability = SystemStabilityKind.STABLE
        elif num_unstable > num_forwards:
            self.system_stability = SystemStabilityKind.NO_STABLE
        else:
            self.system_stability = SystemStabilityKind.MULTIPLE_STABLE

    def _classify_transition_vector_stability(
        self,
        /,
        tolerance: float = 1e-12,
    ) -> None:
        self.transition_vector_stability \
            = _classify_solution_vector_stability(
                self.Ua,
                self.num_unit_roots,
                tolerance=tolerance,
            )

    def _classify_measurement_vector_stability(
        self,
        /,
        tolerance: float = 1e-12,
    ) -> None:
        self.measurement_vector_stability \
            = _classify_solution_vector_stability(
                self.Za,
                self.num_unit_roots,
                tolerance=tolerance,
            )
    #]


def _left_div(A, B):
    r"""
    Solve A \ B
    """
    return _np.linalg.lstsq(A, B, rcond=None)[0]


def _right_div(B, A):
    r"""
    Solve B / A which is (A' \ B')'
    """
    return _np.linalg.lstsq(A.T, B.T, rcond=None)[0].T


def _square_from_triangular(
    triangular_solution: tuple[_np.ndarray, ...],
    /,
) -> tuple[_np.ndarray, ...]:
    r"""
    T <- Ua • Ta / Ua
    R <- Ua • Ra
    X <- Xa • Ra
    K <- Ua • Ka
    xi[t] = ... -X J**(k-1) Ru e[t+k]
    """
    #[
    Ua, Ta, Ra, Ka, Xa, *_ = triangular_solution
    T = Ua @ _right_div(Ta, Ua) # Ua • Ta • inv(Ua)
    R = Ua @ Ra
    K = Ua @ Ka
    X = Ua @ Xa
    return T, R, K, X
    #]


def detach_stable_from_unit_roots(
    transition_solution_prelim: tuple[_np.ndarray, ...],
    is_unit_root: Callable[[Real], bool],
    /,
) -> tuple[_np.ndarray, ...]:
    """
    Apply a secondary Schur decomposition to detach stable eigenvalues from unit roots
    """
    #[
    Ug, Tg, Rg, Kg, Xg, J, Ru = transition_solution_prelim
    num_xib = Tg.shape[0]
    Ta, u, check_num_unit_roots = _sp.linalg.schur(Tg, sort=is_unit_root, ) # Tg = u @ Ta @ u.T
    Ua = Ug @ u if Ug is not None else u
    Ra = u.T @ Rg
    Ka = u.T @ Kg
    Xa = u.T @ Xg if Xg is not None else None
    return Ua, Ta, Ra, Ka, Xa, J, Ru
    #]


def _solve_measurement_equations(descriptor, system, Ua, ) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    num_forwards = descriptor.get_num_forwards()
    G = system.G[:, num_forwards:]
    Z = _left_div(-system.F, G) # -F \ G
    H = _left_div(-system.F, system.J) # -F \ J
    D = _left_div(-system.F, system.H) # -F \ H
    Za = Z @ Ua
    return Z, H, D, Za
    #]


def _solve_transition_equations(
    descriptor,
    system,
    qz,
    /,
) -> tuple[_np.ndarray, ...]:
    """
    """
    #[
    num_backwards = descriptor.get_num_backwards()
    num_forwards = descriptor.get_num_forwards()
    num_stable = num_backwards
    S, T, Q, Z = qz
    A, B, C, D = system.A, system.B, system.C, system.D
    #
    S11 = S[:num_stable, :num_stable]
    S12 = S[:num_stable, num_stable:]
    S22 = S[num_stable:, num_stable:]
    #
    T11 = T[:num_stable, :num_stable]
    T12 = T[:num_stable, num_stable:]
    T22 = T[num_stable:, num_stable:]
    #
    Z21 = Z[num_forwards:, :num_stable]
    Z22 = Z[num_forwards:, num_stable:]
    #
    QC = Q @ C
    QC1 = QC[:num_stable, ...]
    QC2 = QC[num_stable:, ...]
    #
    QD = Q @ D
    QD1 = QD[:num_stable, ...]
    QD2 = QD[num_stable:, ...]
    #
    # Unstable block
    #
    G = _left_div(-Z21, Z22) # -Z21 \ Z22
    Ru = _left_div(-T22, QD2) # -T22 \ QD2
    Ku = _left_div(-(S22 + T22), QC2) # -(S22+T22) \ QC2
    #
    # Transform stable block==transform backward-looking variables:
    # gamma(t) = s(t) + G u(t+1)
    #
    Xg0 = _left_div(S11, T11 @ G + T12)
    Xg1 = G + _left_div(S11, S12)
    #
    Tg = _left_div(-S11, T11)
    Rg = -Xg0 @ Ru - _left_div(S11, QD1)
    Kg = -(Xg0 + Xg1) @ Ku - _left_div(S11, QC1)
    Ug = Z21 # xib = Ug @ gamma
    #
    # Forward expansion
    # gamma(t) = ... -Xg J**(k-1) Ru e(t+k)
    #
    J = _left_div(-T22, S22) # -T22 \ S22
    Xg = Xg1 + Xg0 @ J
    #
    return Ug, Tg, Rg, Kg, Xg, J, Ru
    #]


def _solve_ordqz(
    system,
    is_alpha_beta_stable_or_unit_root,
    is_stable_root,
    is_unit_root,
    /,
) -> tuple[tuple[_np.ndarray, ...], tuple[Real, ], ]:
    """
    """
    #[
    S, T, alpha, beta, Q, Z = _sp.linalg.ordqz(system.A, system.B, sort=is_alpha_beta_stable_or_unit_root, )
    Q = Q.T
    #
    inx_nonzero_alpha = alpha != 0
    eigen_values = _np.full(beta.shape, _np.inf, dtype=complex, )
    eigen_values[inx_nonzero_alpha] = -beta[inx_nonzero_alpha] / alpha[inx_nonzero_alpha]
    eigen_values = tuple(eigen_values)
    #
    #
    return (S, T, Q, Z), eigen_values
    #]


def _classify_eigen_value_stability(
    eigen_value,
    is_stable_root,
    is_unit_root,
) -> EigenValueKind:
    #[
    abs_eigen_value = _np.abs(eigen_value)
    if is_stable_root(abs_eigen_value):
        return EigenValueKind.STABLE
    elif is_unit_root(abs_eigen_value):
        return EigenValueKind.UNIT
    else:
        return EigenValueKind.UNSTABLE
    #]


def _classify_solution_vector_stability(
    transform_matrix: _np.ndarray,
    num_unit_roots: int,
    /,
    tolerance: float = 1e-12,
) -> None:
    """
    """
    #[
    test_matrix = _np.abs(transform_matrix[:, :num_unit_roots], )
    index = _np.any(test_matrix > tolerance, axis=1, )
    return tuple(
        VariableStability.UNIT_ROOT if i
        else VariableStability.STABLE
        for i in index
    )
    #]




"""
Time series chartpacks
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )
import re as _re
#]


class Chartpack:
    """
    """
    #[

    __slots__ = (
        "transforms",
        "charts",
    )

    #]


class Chart:
    """
    """
    #[

    _INPUT_STRING_PATTERN = _re.compile(r"^(?P<title>.*?:)(?P<expression>.*?)(?:\[(?P<transform>.*?)\])?$")

    __slots__ = (
        "title",
        "expression",
        "transform"
    )

    def __init__(
        self,
        expression: str,
        title: str | None = None,
        transform: str | Callable | None = None,
    ) -> None:
        """
        """
        self.expression = expression
        self.title = title or None
        self.transform = transform or None

    @classmethod
    def from_string(
        klass: type(Self),
        input_string: str,
    ) -> Self:
    pass

    #]

    + ["x", "inflation: pct_cpi [dev]"



"""
Plotly interface to time series objects
"""


#[
from __future__ import annotations

from typing import (Self, Sequence, Iterable, Any, )
from types import EllipsisType
import os as _os
import json as _js
import copy as _cp
import plotly.graph_objects as _pg
import itertools as _it

from .. import dates as _dates
#]


_PLOTLY_STYLES_FOLDER = _os.path.join(_os.path.dirname(__file__), "plotly_styles")
_PLOTLY_STYLES = {
    "layouts": {},
}
with open(_os.path.join(_PLOTLY_STYLES_FOLDER, "plain_layout.json", ), "rt", ) as fid:
    _PLOTLY_STYLES["layouts"]["plain"] = _js.load(fid, )


def _line_plot(color: str, **settings) -> _pg.Scatter:
    """
    """
    settings = {"mode": "lines+markers", } | settings
    return _pg.Scatter(line_color=color, **settings, )


def _bar_plot(color: str, **settings) -> _pg.Bar:
    """
    """
    return _pg.Bar(marker_color=color, **settings, )


_PLOTLY_TRACES_FACTORY = {
    "line": _line_plot,
    "bar": _bar_plot,
}


_COLOR_ORDER = [
  "rgb(  0,   114,   189)",
  "rgb(217,    83,    25)",
  "rgb(237,   177,    32)",
  "rgb(126,    47,   142)",
  "rgb(119,   172,    48)",
  "rgb( 77,   190,   238)",
  "rgb(162,    20,    47)",
]


class Mixin:
    """
    """
    #[

    def plot(
        self,
        *,
        span: Iterable[_dates.Dater] | EllipsisType = ...,
        figure_title: str | None = None,
        subplot_title: str | None = None,
        legend: Iterable[str] | None = None,
        show_figure: bool = False,
        show_legend: bool | None = None,
        figure = None,
        subplot: tuple[int, int] | int | None = None,
        xline = None,
        type: Literal["line", "bar"] = "line",
        traces: tuple(dict[str, Any], ) | None = None,
        layout: dict[str, Any] | None = None,
    ) -> _pg.Figure:
        """
        """
        span = self._resolve_dates(span, )
        span = [ t for t in span ]
        frequency = span[0].frequency
        data = self.get_data(span, )
        num_variants = data.shape[1]
        date_str = [ t.to_plotly_date() for t in span ]
        date_format = span[0].frequency.plotly_format
        figure = _pg.Figure() if figure is None else figure
        tile, index = _resolve_subplot(figure, subplot, )
        show_legend = show_legend if show_legend is not None else legend is not None

        minor_dtick_string = None
        if frequency.is_regular:
            minor_dtick_months = min(12//frequency, 1, )
            minor_dtick_string = f"M{minor_dtick_months}"

        traces = (traces, ) if isinstance(traces, dict) else traces
        color_cycle = _it.cycle(_COLOR_ORDER)
        traces_cycle = _it.cycle(traces or ({}, ))

        for i, c, ts in zip(range(num_variants, ), color_cycle, traces_cycle, ):
            traces_settings = {
                "x": date_str,
                "y": data[:, i],
                "name": legend[i] if legend else None,
                "showlegend": show_legend,
                "xhoverformat": date_format,
            }
            traces_settings |= ts or {}
            traces_object = _PLOTLY_TRACES_FACTORY[type](c, **traces_settings, )
            figure.add_trace(traces_object, row=tile[0]+1, col=tile[1]+1, )

        # REFACTOR
        xaxis = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"]["xaxis"])
        yaxis = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"]["yaxis"])
        layout = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"])
        del layout["xaxis"]
        del layout["yaxis"]

        xaxis["tickformat"] = date_format
        xaxis["ticklabelmode"] = "period"
        xaxis["minor"] = {"showgrid": True, "dtick": minor_dtick_string, }
        figure.update_xaxes(xaxis, row=tile[0]+1, col=tile[1]+1, )
        figure.update_layout(title={"text": figure_title}, )
        figure.update_layout(layout or {}, )

        if subplot_title:
            _update_subplot_title(figure, subplot_title, index, )

        if xline:
            xline = (
                xline.to_plotly_date()
                if hasattr(xline, "to_plotly_date")
                else xline
            )
            figure.add_vline(xline, )

        if show_figure:
            figure.show()

        return figure

    #]


def _resolve_subplot(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
    /,
) -> tuple[tuple[int, int], int]:
    """
    """
    rows, columns = figure._get_subplot_rows_columns()
    num_rows = len(rows)
    num_columns = len(columns)
    if subplot is None:
        tile = None
        index = None
    elif isinstance(subplot, Sequence):
        row = subplot[0]
        column = subplot[1]
        index = row * num_columns + column
        tile = row, column,
    elif isinstance(subplot, int):
        index = subplot
        row = index // num_columns
        column = index % num_columns
        tile = row, column,
    else:
        raise TypeError(f"Invalid subplot type: {type(subplot)}")
    return tile, index


def _update_subplot_title(
    figure: _pg.Figure,
    subplot_title: str,
    index: int,
    /,
) -> None:
    """
    """
    annotation = next(figure.select_annotations(index, ), None, )
    if annotation:
        annotation.text = subplot_title

